{-# LANGUAGE TemplateHaskell #-}

import qualified Buffer as B
import qualified Keymap as KM
import qualified VtyUI as UI
import qualified Widget as W

import Control.Arrow (first)
import Control.Exception
import Control.Lens
import Control.Monad.RWS
import qualified Data.Array as A
import qualified Data.Map as M
import Data.Maybe
import qualified Data.Text.Lazy as TL
import Graphics.Vty
import System.Environment (getArgs)
import System.IO.Error
import Text.Printf (printf)
import Text.Read (readMaybe)
import Text.Regex.PCRE
import qualified Yi.Rope as R

data EditorMode = Command
                | Editing
                | Selection
                | Interactive
                | Error
                deriving (Eq)

data SelectionMode = CharWise
                   | LineWise
                   deriving (Eq)

data SelectionState =
  SelectionState { sMode   :: SelectionMode
                 , sAnchor :: B.Anchor
                 } deriving (Eq)

data Direction = Forward | Backward deriving (Eq)

data SearchContext = SearchContext { _scFrom      :: B.Point
                                   , _scDirection :: Direction
                                   }

data Options = Options
  { _optKeyboardLayout         :: KM.KeyboardLayout
  , _optEnableSmartNewline     :: Bool
  , _optEnableSmartStartOfLine :: Bool
  , _optShowLineNumbers        :: Bool
  }

defaultOptions :: Options
defaultOptions =
  Options { _optKeyboardLayout         = KM.Dvorak
          , _optEnableSmartNewline     = True
          , _optEnableSmartStartOfLine = True
          , _optShowLineNumbers        = True
          }
keyboardLayout :: KM.KeyboardLayout
keyboardLayout = KM.Dvorak

data EditorState =
  EditorState { _esBufferZipper  :: B.BufferZipper
              , esSelection      :: Maybe SelectionState
              , _esMode          :: EditorMode
              , esClipboard      :: Maybe R.YiString
              , _esOptions       :: Options
              , esKeyPrefix      :: Maybe Event
              , esLastEvent      :: Maybe Event -- TODO: Merge prefix and last event?
              , esDialog         :: Maybe Dialog
              , _esSearchContext :: SearchContext
              , _esError         :: Maybe SomeException
              , esQuit           :: Bool
              }

type Editor = RWST Vty () EditorState IO

data Dialog = Dialog { dTag    :: String
                     -- ^ Tag (ID) that is displayed on the status line.
                     , dPrompt :: TL.Text
                     , dText   :: TL.Text
                     , dPoint  :: B.Point
                     -- ^ The point of the currently active buffer when initiating the dialog.
                     , dUpdate :: Maybe (Editor ())
                     , dFinish :: Maybe (Editor ())
                     }

makeLenses ''SearchContext
makeLenses ''EditorState
makeLenses ''Options

commandModeKMs :: [KM.Keymap (Editor ())]
commandModeKMs = [movementCommonKM, movementKM, jumpKM, openKM, changeKM, deletionCommonKM, deletionKM, topLevelKM]

-- Common keymap that is not dependent on a keyboard layout.
movementCommonKM :: KM.Keymap (Editor ())
movementCommonKM = KM.Keymap Nothing $ M.fromList
  [ (EvKey KRight [], right)
  , (EvKey KLeft  [], left)
  , (EvKey KDown  [], down)
  , (EvKey KUp    [], up)
  , (EvKey KEnd   [], eol)
  , (EvKey KHome  [], bolSmart)
  , (EvKey KPageDown [], pageDown)
  , (EvKey KPageUp   [], pageUp)
  , (EvKey (KFun 2) [], toggleShowLineNumbers)
  , (EvKey (KFun 3) [], toggleEnableSmartNewline)
  , (EvKey (KFun 4) [], toggleEnableSmartStartOfLine)
  ]

movementKM :: KM.Keymap (Editor ())
movementKM = if keyboardLayout == KM.Dvorak then dvorak else qwerty
  where
    dvorak = KM.Keymap Nothing $ M.fromList
      [ (EvKey (KChar 'n') [], right)
      , (EvKey (KChar 'h') [], left)
      , (EvKey (KChar 't') [], down)
      , (EvKey (KChar 'c') [], up)
      , (EvKey (KChar 'r') [], wordRight)
      , (EvKey (KChar 'g') [], wordLeft)
      , (EvKey (KChar 'N') [], eol)
      , (EvKey (KChar 'H') [], bolSmart)
      , (EvKey (KChar 'T') [], pageDown)
      , (EvKey (KChar 'C') [], pageUp)
      , (EvKey (KChar 'w') [], leaveMark)
      , (EvKey (KChar 'p') [], delete)
      , (EvKey (KChar '.') [], backspace)
      , (EvKey (KChar 'P') [], deleteWord)
      , (EvKey (KChar '>') [], backspaceWord)
      , (EvKey (KChar ',') [], cleanupWhitespace)
      -- Ctrl-@ is identical to Ctrl-<space>.
      -- , (EvKey (KChar '@') [MCtrl], selectionMode LineWise)
      , (EvKey (KChar 'q') [], cut)
      , (EvKey (KChar 'j') [], copy)
      , (EvKey (KChar 'k') [], paste)
      , (EvKey (KChar 'z') [], undo)
      , (EvKey (KChar 'Z') [], redo)
      , (EvKey (KChar 's') [], search Forward)
      , (EvKey (KChar 'S') [], search Backward)
      , (EvKey (KChar ' ') [], selectionMode)
      ]
    qwerty = KM.Keymap Nothing $ M.fromList
      [ (EvKey (KChar 'l') [], right)
      , (EvKey (KChar 'j') [], left)
      , (EvKey (KChar 'k') [], down)
      , (EvKey (KChar 'i') [], up)
      , (EvKey (KChar 'o') [], wordRight)
      , (EvKey (KChar 'u') [], wordLeft)
      , (EvKey (KChar 'L') [], eol)
      , (EvKey (KChar 'J') [], bolSmart)
      , (EvKey (KChar 'K') [], pageDown)
      , (EvKey (KChar 'I') [], pageUp)
      , (EvKey (KChar ',') [], leaveMark)
      , (EvKey (KChar 'r') [], delete)
      , (EvKey (KChar 'e') [], backspace)
      , (EvKey (KChar 'R') [], deleteWord)
      , (EvKey (KChar 'E') [], backspaceWord)
      , (EvKey (KChar 'w') [], cleanupWhitespace)
      , (EvKey (KChar 'x') [], cut)
      , (EvKey (KChar 'c') [], copy)
      , (EvKey (KChar 'v') [], paste)
      , (EvKey (KChar '/') [], undo)
      , (EvKey (KChar '?') [], redo)
      , (EvKey (KChar ';') [], search Forward)
      , (EvKey (KChar ':') [], search Backward)
      , (EvKey (KChar ' ') [], selectionMode)
      ]

jumpKM :: KM.Keymap (Editor ())
jumpKM = if keyboardLayout == KM.Dvorak then dvorak else qwerty
  where
    dvorak = KM.Keymap (Just $ EvKey (KChar 'm') []) $ M.fromList
      [ (EvKey (KChar 't') [], eob)
      , (EvKey (KChar 'c') [], bob)
      , (EvKey (KChar 'm') [], gotoLineInteractive)
      , (EvKey (KChar 'r') [], gotoLastEditEnd)
      , (EvKey (KChar 'g') [], gotoLastEditStart)
      , (EvKey (KChar 'w') [], gotoMark)
      ]
    qwerty = KM.Keymap (Just $ EvKey (KChar 'm') []) $ M.fromList
      [ (EvKey (KChar 'k') [], eob)
      , (EvKey (KChar 'i') [], bob)
      , (EvKey (KChar 'm') [], gotoLineInteractive)
      , (EvKey (KChar 'o') [], gotoLastEditEnd)
      , (EvKey (KChar 'u') [], gotoLastEditStart)
      , (EvKey (KChar ',') [], gotoMark)
      ]


openKM :: KM.Keymap (Editor ())
openKM = if keyboardLayout == KM.Dvorak then dvorak else qwerty
  where
    dvorak = KM.Keymap (Just $ EvKey (KChar 'u') []) $ M.fromList
      [ (EvKey (KChar 'u') [], editingMode)
      , (EvKey (KChar 'n') [], openRight)
      , (EvKey (KChar 'h') [], openLeft)
      , (EvKey (KChar 't') [], openLineBelow)
      , (EvKey (KChar 'c') [], openLineAbove)
      , (EvKey (KChar 'N') [], openEndOfLine)
      , (EvKey (KChar 'H') [], openStartOfLine)
      ]
    qwerty = KM.Keymap (Just $ EvKey (KChar 'f') []) $ M.fromList
      [ (EvKey (KChar 'f') [], editingMode)
      , (EvKey (KChar 'l') [], openRight)
      , (EvKey (KChar 'j') [], openLeft)
      , (EvKey (KChar 'k') [], openLineBelow)
      , (EvKey (KChar 'i') [], openLineAbove)
      , (EvKey (KChar 'L') [], openEndOfLine)
      , (EvKey (KChar 'J') [], openStartOfLine)
      ]

changeKM :: KM.Keymap (Editor ())
changeKM = if keyboardLayout == KM.Dvorak then dvorak else qwerty
  where
    dvorak = KM.Keymap (Just $ EvKey (KChar 'o') []) $ M.fromList
      [ (EvKey (KChar 'n') [], changeCharForward)
      , (EvKey (KChar 'h') [], changeCharBackward)
      , (EvKey (KChar 'N') [], changeEndOfLine)
      , (EvKey (KChar 'H') [], changeStartOfLine)
      , (EvKey (KChar 'o') [], changeLine)
      , (EvKey (KChar '(') [], changeAroundPair '(' ')')
      , (EvKey (KChar ')') [], changeAroundPair '(' ')')
      , (EvKey (KChar '[') [], changeAroundPair '[' ']')
      , (EvKey (KChar ']') [], changeAroundPair '[' ']')
      , (EvKey (KChar '{') [], changeAroundPair '{' '}')
      , (EvKey (KChar '{') [], changeAroundPair '{' '}')
      , (EvKey (KChar '"') [], changeAroundPair '"' '"')
      , (EvKey (KChar '\'') [], changeAroundPair '\'' '\'')
      ]
    qwerty = KM.Keymap (Just $ EvKey (KChar 's') []) $ M.fromList
      [ (EvKey (KChar 'l') [], changeCharForward)
      , (EvKey (KChar 'j') [], changeCharBackward)
      , (EvKey (KChar 'L') [], changeEndOfLine)
      , (EvKey (KChar 'J') [], changeStartOfLine)
      , (EvKey (KChar 's') [], changeLine)
      , (EvKey (KChar '(') [], changeAroundPair '(' ')')
      , (EvKey (KChar ')') [], changeAroundPair '(' ')')
      , (EvKey (KChar '[') [], changeAroundPair '[' ']')
      , (EvKey (KChar ']') [], changeAroundPair '[' ']')
      , (EvKey (KChar '{') [], changeAroundPair '{' '}')
      , (EvKey (KChar '{') [], changeAroundPair '{' '}')
      , (EvKey (KChar '"') [], changeAroundPair '"' '"')
      , (EvKey (KChar '\'') [], changeAroundPair '\'' '\'')
      ]

deletionCommonKM :: KM.Keymap (Editor ())
deletionCommonKM = KM.Keymap Nothing $ M.fromList
  [ (EvKey KDel [], delete)
  , (EvKey KBS  [], backspace)
  ]

deletionKM :: KM.Keymap (Editor ())
deletionKM = if keyboardLayout == KM.Dvorak then dvorak else qwerty
  where
    dvorak = KM.Keymap (Just $ EvKey (KChar 'e') []) $ M.fromList
      [ (EvKey (KChar 'N') [], deleteEndOfLine)
      , (EvKey (KChar 'H') [], deleteStartOfLine)
      , (EvKey (KChar 'e') [], deleteLine)
      , (EvKey (KChar '(') [], deleteAroundPair '(' ')')
      , (EvKey (KChar ')') [], deleteAroundPair '(' ')')
      , (EvKey (KChar '[') [], deleteAroundPair '[' ']')
      , (EvKey (KChar ']') [], deleteAroundPair '[' ']')
      , (EvKey (KChar '{') [], deleteAroundPair '{' '}')
      , (EvKey (KChar '{') [], deleteAroundPair '{' '}')
      , (EvKey (KChar '"') [], deleteAroundPair '"' '"')
      , (EvKey (KChar '\'') [], deleteAroundPair '\'' '\'')
      ]
    qwerty = KM.Keymap (Just $ EvKey (KChar 'd') []) $ M.fromList
      [ (EvKey (KChar 'L') [], deleteEndOfLine)
      , (EvKey (KChar 'J') [], deleteStartOfLine)
      , (EvKey (KChar 'd') [], deleteLine)
      , (EvKey (KChar '(') [], deleteAroundPair '(' ')')
      , (EvKey (KChar ')') [], deleteAroundPair '(' ')')
      , (EvKey (KChar '[') [], deleteAroundPair '[' ']')
      , (EvKey (KChar ']') [], deleteAroundPair '[' ']')
      , (EvKey (KChar '{') [], deleteAroundPair '{' '}')
      , (EvKey (KChar '{') [], deleteAroundPair '{' '}')
      , (EvKey (KChar '"') [], deleteAroundPair '"' '"')
      , (EvKey (KChar '\'') [], deleteAroundPair '\'' '\'')
      ]


-- TODO: Decide whether Enter is a good key to use to enter command mode.
topLevelKM :: KM.Keymap (Editor ())
topLevelKM = KM.Keymap Nothing $ M.fromList
  [ (EvKey (KChar 'l') [MCtrl], redisplay)
  , (EvKey KRight [MCtrl], nextBuffer)
  , (EvKey KLeft  [MCtrl], previousBuffer)
  , (EvKey (KChar 's') [MCtrl], save)
  , (EvKey (KChar 'o') [MCtrl], load)
  , (EvKey (KChar 'k') [MCtrl], close)
  , (EvKey (KChar 'q') [MCtrl], quit)
  ]

editingKM :: KM.Keymap (Editor ())
editingKM = KM.Keymap Nothing $ M.fromList
  [ (EvKey KRight [], right)
  , (EvKey KLeft  [], left)
  , (EvKey KDown  [], down)
  , (EvKey KUp    [], up)
  , (EvKey KPageDown [], pageDown)
  , (EvKey KPageUp   [], pageUp)
  , (EvKey KDel [], delete)
  , (EvKey KBS  [], backspace)
  , (EvKey KEsc [], commandMode)
  , (EvKey (KChar 'm') [MMeta], commandMode)
  , (EvKey (KChar 'q') [MCtrl], quit)
  ]

selectionModeKM :: KM.Keymap (Editor ())
selectionModeKM = if keyboardLayout == KM.Dvorak then dvorak else qwerty
  where
    dvorak = KM.Keymap Nothing $ M.fromList
      [ (EvKey (KChar 'n') [], right)
      , (EvKey (KChar 'h') [], left)
      , (EvKey (KChar 't') [], down)
      , (EvKey (KChar 'c') [], up)
      , (EvKey (KChar 'r') [], wordRight)
      , (EvKey (KChar 'g') [], wordLeft)
      , (EvKey (KChar 'N') [], eol)
      , (EvKey (KChar 'H') [], bolSmart)
      , (EvKey (KChar 'T') [], pageDown)
      , (EvKey (KChar 'C') [], pageUp)
      , (EvKey (KChar '(') [], wrapSelection "(" ")")
      , (EvKey (KChar ')') [], wrapSelection "(" ")")
      , (EvKey (KChar '[') [], wrapSelection "[" "]")
      , (EvKey (KChar ']') [], wrapSelection "[" "]")
      , (EvKey (KChar '{') [], wrapSelection "{" "}")
      , (EvKey (KChar '}') [], wrapSelection "{" "}")
      , (EvKey (KChar '"') [], wrapSelection "\"" "\"")
      , (EvKey (KChar '\'') [], wrapSelection "'" "'")
      , (EvKey (KChar 'o') [], changeSelection)
      , (EvKey (KChar 'q') [], cut)
      , (EvKey (KChar 'j') [], copy)
      , (EvKey (KChar 'k') [], paste)
      , (EvKey (KChar 'a') [], selectionSwapAnchor)
      , (EvKey (KChar ' ') [], toggleSelectionMode)
      , (EvKey KEsc        [], commandMode)
      ] `M.union` (KM.kmBindings movementCommonKM)

    qwerty = KM.Keymap Nothing $ M.fromList
      [ (EvKey (KChar 'l') [], right)
      , (EvKey (KChar 'j') [], left)
      , (EvKey (KChar 'k') [], down)
      , (EvKey (KChar 'i') [], up)
      , (EvKey (KChar 'o') [], wordRight)
      , (EvKey (KChar 'u') [], wordLeft)
      , (EvKey (KChar 'L') [], eol)
      , (EvKey (KChar 'J') [], bolSmart)
      , (EvKey (KChar 'K') [], pageDown)
      , (EvKey (KChar 'I') [], pageUp)
      , (EvKey (KChar '(') [], wrapSelection "(" ")")
      , (EvKey (KChar ')') [], wrapSelection "(" ")")
      , (EvKey (KChar '[') [], wrapSelection "[" "]")
      , (EvKey (KChar ']') [], wrapSelection "[" "]")
      , (EvKey (KChar '{') [], wrapSelection "{" "}")
      , (EvKey (KChar '}') [], wrapSelection "{" "}")
      , (EvKey (KChar '"') [], wrapSelection "\"" "\"")
      , (EvKey (KChar '\'') [], wrapSelection "'" "'")
      , (EvKey (KChar 's') [], changeSelection)
      , (EvKey (KChar 'x') [], cut)
      , (EvKey (KChar 'c') [], copy)
      , (EvKey (KChar 'v') [], paste)
      , (EvKey (KChar 'a') [], selectionSwapAnchor)
      , (EvKey (KChar ' ') [], toggleSelectionMode)
      , (EvKey KEsc        [], commandMode)
      ] `M.union` (KM.kmBindings movementCommonKM)

editorBuffer es = view (esBufferZipper . B.bufCurrent) es
updateBuffer = assign $ esBufferZipper . B.bufCurrent
startEdit, endEdit :: Editor ()
startEdit = get >>= updateBuffer . B.startEdit . editorBuffer
endEdit = get >>= updateBuffer . B.endEdit . editorBuffer

move :: (B.Buffer -> B.Buffer) -> Editor ()
move action = do
  es <- get
  updateBuffer $ B.adjustWindow $ action $ editorBuffer es

right :: Editor ()
right = move B.moveRight

left :: Editor ()
left = move B.moveLeft

down :: Editor ()
down = move B.moveDown

up :: Editor ()
up = move B.moveUp

wordRight :: Editor()
wordRight = move B.moveWordRight

wordLeft :: Editor()
wordLeft = move B.moveWordLeft

eol :: Editor ()
eol = move B.moveEndOfLine

bol :: Editor ()
bol = move $ B.moveStartOfLine False

bolSmart :: Editor ()
bolSmart = get >>= move . B.moveStartOfLine . view (esOptions . optEnableSmartStartOfLine)

pageDown :: Editor ()
pageDown = move B.movePageDown

pageUp :: Editor ()
pageUp = move B.movePageUp

eob :: Editor ()
eob = move B.moveEndOfText

bob :: Editor ()
bob = move B.moveStartOfText

gotoPosition :: Int -> Editor ()
gotoPosition = move . B.gotoPosition

gotoLine :: Int -> Editor ()
gotoLine = move . B.gotoLine

gotoLastEdit :: Bool -> Editor ()
gotoLastEdit start = do
  es <- get
  case B.lastEditPoint $ editorBuffer es of
    Nothing     -> return ()
    Just (s, e) -> gotoPosition $ if start then s else e

gotoLastEditEnd :: Editor ()
gotoLastEditEnd = gotoLastEdit False

gotoLastEditStart :: Editor ()
gotoLastEditStart = gotoLastEdit True

leaveMark :: Editor ()
leaveMark = get >>= updateBuffer . B.leaveMark . editorBuffer

gotoMark :: Editor ()
gotoMark = get >>= gotoPosition . B.mark . editorBuffer

openRight :: Editor ()
openRight = right >> editingMode

openLeft :: Editor ()
openLeft = left >> editingMode

openLineBelow :: Editor ()
openLineBelow = eol >> insertSmartNewline >> editingMode

openLineAbove :: Editor ()
openLineAbove = bol >> left >> insertSmartNewline >> editingMode

openEndOfLine :: Editor ()
openEndOfLine = eol >> editingMode

openStartOfLine :: Editor ()
openStartOfLine = bolSmart >> editingMode

changeCharForward :: Editor ()
changeCharForward = delete >> editingMode

changeCharBackward :: Editor ()
changeCharBackward = backspace >> editingMode

changeEndOfLine :: Editor ()
changeEndOfLine = deleteEndOfLine >> editingMode

changeStartOfLine :: Editor ()
changeStartOfLine = deleteStartOfLine >> editingMode

changeLine :: Editor ()
changeLine = bol >> deleteEndOfLine >> editingMode

changeAroundPair :: Char -> Char -> Editor ()
changeAroundPair left right = deleteAroundPair left right >> editingMode

edit :: (B.Buffer -> B.Buffer) -> Editor ()
edit action = do
  assign (esBufferZipper . B.bufCurrent . B.isModified) True
  es <- get
  updateBuffer $ B.adjustWindow $ action $ editorBuffer es

insertText :: String -> Editor ()
insertText txt = do
  edit $ B.insertTextWithUndo $ R.fromString txt

insertSmartNewline :: Editor()
insertSmartNewline = do
  es <- get
  let i = B.lineIndent $ editorBuffer es
  insertText $ "\n" ++ indent i es
  where
    indent i es = if view (esOptions . optEnableSmartNewline) es then replicate i ' ' else ""

delete :: Editor ()
delete = edit $ B.deleteTextWithUndo 1

backspace :: Editor ()
backspace = do
  es <- get
  let b = editorBuffer es
  when (B.point b > 0) $ do
    edit $ B.deleteTextWithUndo 1 . B.moveLeft

deleteWord :: Editor ()
deleteWord = do
  es <- get
  let b = editorBuffer es
  let p = B.point $ B.moveWordRight b
  edit $ B.deleteTextWithUndo (p - B.point b)

backspaceWord :: Editor ()
backspaceWord = do
  es <- get
  let p = B.point $ editorBuffer es
  move B.moveWordLeft
  es <- get
  edit $ B.deleteTextWithUndo (p - B.point (editorBuffer es))

deleteEndOfLine :: Editor ()
deleteEndOfLine = do
  es <- get
  let b = editorBuffer es
  edit $ B.deleteTextWithUndo (B.endOfLine b - B.point b)
  updateClipboard

-- TODO: It would be nicer to support this directly in Buffer.
deleteStartOfLine :: Editor ()
deleteStartOfLine = do
  es <- get
  let p = B.point $ editorBuffer es
  bol
  es <- get
  let b =  editorBuffer es
  edit $ B.deleteTextWithUndo (p - B.point b)
  updateClipboard

deleteLine :: Editor ()
deleteLine = do
  bol
  es <- get
  let b = editorBuffer es
  edit $ B.deleteTextWithUndo (B.endOfLine b - B.startOfLine b + 1)
  updateClipboard

deleteAroundPair :: Char -> Char -> Editor ()
deleteAroundPair left right = do
  es <- get
  let b = editorBuffer es
  case B.findPair left right b of
    Nothing -> return ()
    Just (start, end) -> do
      -- TODO: Duplicate code, see deleteSelection.
      move $ B.gotoPosition start
      es <- get
      updateBuffer $ B.deleteTextWithUndo (end - start) $ editorBuffer es
      assign (esBufferZipper . B.bufCurrent . B.isModified) True
      updateClipboard

cleanupWhitespace :: Editor ()
cleanupWhitespace = edit B.cleanupWhitespace


-- Update clipboard from the last deleted text.
updateClipboard :: Editor ()
updateClipboard = do
  es <- get
  let txt = B.lastDeletedText $ editorBuffer es
  put $ es { esClipboard = Just txt }

redisplay :: Editor ()
redisplay = do
  es <- get
  updateBuffer $ B.redisplay $ editorBuffer es

switchMode :: EditorMode -> Editor ()
switchMode mode = esMode .= mode

commandMode = do
  es <- get
  let b = editorBuffer es
  let bz = _esBufferZipper es
  put $ es { esSelection = Nothing, _esBufferZipper = bz { B._bufCurrent = B.clearSelections b } }
  endEdit
  switchMode Command

editingMode = startEdit >> switchMode Editing


selectionRange :: SelectionState -> B.Buffer -> (Int, Int)
selectionRange (SelectionState mode anchor) b = (from, to)
  where
    pos    = B.point b
    start  = case mode of
               CharWise -> snd' anchor
               LineWise -> if pos > snd' anchor
                           then fst' anchor
                           else trd' anchor
    end    = case mode of
               CharWise -> pos
               LineWise -> if pos > snd' anchor
                           then B.endOfLine b
                           else B.startOfLine b
    from   = min start end
    to     = max start end

selectionMode :: Editor ()
selectionMode = do
  es <- get
  let s = esSelection es
  put $ es { esSelection = Just $ SelectionState CharWise $ anchor $ editorBuffer es }
  switchMode Selection
  where
    anchor b = (B.startOfLine b, B.point b, B.endOfLine b)

toggleSelectionMode :: Editor ()
toggleSelectionMode = do
  es <- get
  case esSelection es of
    Nothing -> return ()
    Just state@(SelectionState mode _) ->
      put $ es { esSelection = Just $ state { sMode = if mode == CharWise then LineWise else CharWise } }

updateSelection :: Editor ()
updateSelection = do
  es <- get
  let b = editorBuffer es
  let s = fromJust $ esSelection es
  case esSelection es of
    Nothing -> return ()
    Just _  -> updateBuffer $ b { B.selections = [selectionRange s b] }

selectionSwapAnchor :: Editor ()
selectionSwapAnchor = do
  es <- get
  let b = editorBuffer es
  let s = fromJust $ esSelection es
  put $ es { esSelection = Just $ s { sAnchor = anchor b } }
  gotoPosition $ snd' $ sAnchor s
  where
    anchor b = (B.startOfLine b, B.point b, B.endOfLine b)

copySelection :: Editor ()
copySelection = do
  es <- get
  put $ es { esClipboard = sel es }
  where
    sel es = case view esMode es of
               Selection -> Just $ R.take (end + newline - start) $ R.drop start $ B.text b
               otherwise -> Nothing
       where
          s = fromJust $ esSelection es
          b = editorBuffer es
          (start, end) = selectionRange s b
          newline = if sMode s == LineWise then 1 else 0

wrapSelection :: String -> String -> Editor ()
wrapSelection left right = do
  startEdit
  es <- get
  let s = fromJust $ esSelection es
  let (start, end) = selectionRange s $ editorBuffer es
  gotoPosition end
  insertText right
  gotoPosition start
  insertText left
  commandMode

changeSelection :: Editor ()
changeSelection = deleteSelection >> editingMode

deleteSelection :: Editor ()
deleteSelection = do
  es <- get
  let s = fromJust $ esSelection es
  let b = editorBuffer es
  let (start, end) = selectionRange s b
  let newline = if sMode s == LineWise
                then 1
                else 0
  commandMode
  move $ B.gotoPosition start
  es <- get
  updateBuffer $ B.deleteTextWithUndo (end - start + newline) $ editorBuffer es
  assign (esBufferZipper . B.bufCurrent . B.isModified) True
  updateClipboard


cut :: Editor ()
cut = do
  es <- get
  -- Smart cut.
  case view esMode es of
    Selection -> copySelection >> deleteSelection
    otherwise -> deleteLine


copy :: Editor ()
copy = do
  es <- get
  -- Smart copy.
  case view esMode es of
    Selection -> copySelection >> commandMode
    otherwise -> put $ es { esClipboard = Just $ B.currentLine $ editorBuffer es }


paste :: Editor ()
paste = do
  es <- get
  let txt = fromMaybe R.empty (esClipboard es)
  when (view esMode es == Selection) $ do
    deleteSelection
  edit $ B.insertTextWithUndo txt

-- TODO:
---- 1) Backspace is special in that it moves the cursor. Its undo doesn't work as it should right now.
---- 2) It would be nice to store the buffer's modified flag in the undo record and restore it on undo/redo.
undo :: Editor ()
undo = edit B.undo

redo :: Editor ()
redo = edit B.redo

interactive :: String -> String -> Maybe (Editor ()) -> Maybe (Editor ()) -> Editor ()
interactive tag prompt update finish = do
  es <- get
  put $ es { esDialog = Just $ Dialog { dTag    = tag
                                      , dPrompt = TL.pack prompt
                                      , dText   = TL.empty
                                      , dPoint  = B.point $ editorBuffer es
                                      , dUpdate = update
                                      , dFinish = finish
                                      }
           }
  switchMode Interactive

errorDialog :: Exception e => e -> Editor ()
errorDialog e = assign esError (Just $ toException e) >> switchMode Error

gotoLineInteractive :: Editor ()
gotoLineInteractive = interactive "Line" "Go to line: " (Just gotoLineInteractiveUpdate) Nothing

gotoLineInteractiveUpdate :: Editor ()
gotoLineInteractiveUpdate = do
  es <- get
  case esDialog es of
    Nothing -> return ()
    Just d  -> do
      gotoLine $ fromMaybe (dPoint d) $ fmap (subtract 1) $ readMaybe (TL.unpack $ dText d)
      redisplay

search :: Direction -> Editor ()
search d = do
  es <- get
  assign esSearchContext $ SearchContext (B.point $ editorBuffer es) d
  interactive "Find" "Search: " (Just searchUpdate) Nothing

searchUpdate :: Editor ()
searchUpdate = do
  es <- get
  case esDialog es of
    Just d -> case view (esSearchContext . scDirection) es of
      Forward  -> forward es  d $ editorBuffer es
      Backward -> backward es d $ editorBuffer es
    Nothing -> return ()

    where
      forward es d b = do
        let winStart = R.length $ fst $ R.splitAtLine (B.winY $ B.window b) (B.text b)
        let from = view (esSearchContext . scFrom) es
        let txt = R.drop winStart $ B.text b
        let reSearch = makeRegex (TL.unpack $ dText d) :: Regex
        let results  = matchAll reSearch $ R.toString txt
        let selections = fmap (\(o, l) -> (o+winStart, o+winStart+l)) $ fmap (A.!0) results
        updateBuffer $ b { B.selections = selections }
        if results /= []
          then let found = filter (\(o, _) -> o >= from) selections
               in if found /= []
                  then do
                    let p = fst $ head found
                    gotoPosition p
                    assign (esSearchContext . scFrom) p
                  else do
                    -- Wrap search.
                    gotoPosition 0
                    assign (esSearchContext . scFrom) 0
                    searchUpdate
          else gotoPosition $ dPoint d

      backward es d b = do
        let w = B.window b
        let winStart = R.length $ fst $ R.splitAtLine (B.winY w) (B.text b)
        let winEnd = R.length $ fst $ R.splitAtLine (B.winY w + B.winHeight w) (B.text b)
        let from = view (esSearchContext . scFrom) es
        let txt = R.take winEnd $ B.text b
        let reSearch = makeRegex (TL.unpack $ dText d) :: Regex
        let results  = matchAll reSearch $ R.toString txt
        let selections = fmap (\(o, l) -> (o, o+l)) $ fmap (A.!0) results
        updateBuffer $ b { B.selections = selections }
        if results /= []
          then let found = filter (\(o, _) -> o <= from) selections
               in if found /= []
                  then do
                    let p = fst $ last found
                    gotoPosition p
                    assign (esSearchContext . scFrom) p
                  else do
                    -- Wrap search.
                    gotoPosition $ B.pointMax b
                    assign (esSearchContext . scFrom) $ B.pointMax b
                    searchUpdate
          else gotoPosition $ dPoint d

searchNext :: Direction -> Editor ()
searchNext d = do
  case d of
    Forward  -> esSearchContext . scFrom += 1
    Backward -> esSearchContext . scFrom -= 1
  esSearchContext . scDirection .= d
  searchUpdate

nextBuffer, previousBuffer :: Editor ()
nextBuffer = do
  es <- get
  put $ es { _esBufferZipper = B.nextBuffer $ _esBufferZipper es }
previousBuffer = do
  es <- get
  put $ es { _esBufferZipper = B.previousBuffer $ _esBufferZipper es }


toggleEnableSmartNewline :: Editor ()
toggleEnableSmartNewline = esOptions . optEnableSmartNewline %= not

toggleEnableSmartStartOfLine :: Editor ()
toggleEnableSmartStartOfLine = esOptions . optEnableSmartStartOfLine %= not

toggleShowLineNumbers :: Editor ()
toggleShowLineNumbers = esOptions . optShowLineNumbers %= not


save :: Editor ()
save = do
  es <- get
  let b = editorBuffer es
  case B._filePath b of
    Just path -> do
      r <- liftIO $ try $ R.writeFileUsingText path (B.text b)
      case r of
        Left  e -> errorDialog (e :: IOException)
        Right _ -> assign (esBufferZipper . B.bufCurrent . B.isModified) False
    Nothing -> interactive "Save" "Save as: " Nothing (Just saveFinish)

saveFinish :: Editor ()
saveFinish = do
  es <- get
  let path = TL.unpack $ dText $ fromJust $ esDialog es
  r <- liftIO $ try $ R.writeFileUsingText path (B.text $ editorBuffer es)
  case r of
    Left  e -> errorDialog (e :: IOException)
    Right _ -> do
      assign (esBufferZipper . B.bufCurrent . B.isModified) False
      assign (esBufferZipper . B.bufCurrent . B.filePath) $ Just path

load :: Editor ()
load = do
  es <- get
  interactive "Load" "Open: " Nothing (Just loadFinish)

loadFinish :: Editor ()
loadFinish = do
  es <- get
  let path = TL.unpack $ dText $ fromJust $ esDialog es
  r  <- liftIO $ try $ B.loadBuffer path
  case r of
    Left  e -> if isDoesNotExistError e
               then put $ es { _esBufferZipper = B.openBuffer (B.newBuffer { B.name = path }) $ _esBufferZipper es }
               else errorDialog e
    Right b -> put $ es { _esBufferZipper = B.openBuffer b $ _esBufferZipper es }

close :: Editor ()
close = do
  es <- get
  assign esBufferZipper $ B.closeBuffer $ view esBufferZipper es

quit :: Editor ()
quit = do
  es <- get
  if view B.isModified $ editorBuffer es
    then interactive "Quit" "Really quit? " Nothing $ Just finish
    else put $ es { esQuit = True }
  where
    finish = do
      es <- get
      when (isYes (TL.unpack $ TL.toLower $ dText $ fromJust $ esDialog es)) $ do
        put $ es { esQuit = True }
    isYes txt = txt == "yes" || txt == "y"



dialogInsertText :: String -> Editor ()
dialogInsertText txt = do
  es <- get
  case esDialog es of
    Just d  -> put $ es { esDialog = Just d { dText = dText d `TL.append` TL.pack txt } }
    Nothing -> return ()

dialogBackspace :: Editor ()
dialogBackspace = do
  es <- get
  case esDialog es of
    Just d  -> put $ es { esDialog = Just d { dText = TL.dropEnd 1 $ dText d } }
    Nothing -> return ()


dialogHide :: Editor ()
dialogHide = do
  es <- get
  put $ es { esDialog = Nothing }

-- TODO: This is way too slow. Also, it should not be run after every cursor movement.
display :: EditorState -> Editor ()
display es = do
  vty <- ask
  let opts = view esOptions es
  let showLn = _optShowLineNumbers opts
  let lineNumbersImg = if showLn
                       then foldl (<->) emptyImage $
                            -- For some stupid reason, TL.justifyRight needs Int64 and not Int...
                            fmap (text defAttr . TL.justifyRight (fromIntegral nlw) ' ' . TL.pack . show) [winY+1 .. min (nls+1) (winY+winH)]
                       else emptyImage
  let textImg = UI.drawText (R.unlines clampBuffer) off (B.selections buffer) (B.syntax buffer) winW winH
  let statusLineImg = translateY (UI.rawScreenHeight-2) (statusLine opts)
  let messageImg = case view esError es of
                     Nothing -> text defAttr $ case esDialog es of
                                  Just d  -> dPrompt d `TL.append` dText d
                                  Nothing -> TL.empty
                     Just e  -> text defAttr $ TL.pack $ show e
  liftIO $ update vty $ picForLayers [W.renderTestWidget, lineNumbersImg, translateX (gutW showLn) textImg, statusLineImg <-> messageImg]
  displayPoint (gutW showLn) $ editorBuffer es

  where
    buffer = editorBuffer es
    win = B.window buffer
    winY = B.winY win
    winW = B.winWidth win
    winH = B.winHeight win
    (left, rest) = R.splitAtLine winY (B.text buffer)
    off = R.length left
    nls = R.countNewLines $ B.text buffer
    nlw = length $ show nls
    gutW showLn = if showLn
           then nlw+1
           else 0

    -- TODO: For now, long lines are simply clamped to the window width.
    clampBuffer :: [R.YiString]
    clampBuffer = take (B.winHeight win) $ fmap (R.take winW) ls
    ls = R.lines rest

    statusLine opts = mode <|> left <|> spacer <|> right
      where mode = string (attr `withStyle` bold) "["
                 <|> uncurry string m
                 <|> string (attr `withStyle` bold) "] "
            m = first (attr `withStyle` bold `withForeColor`) $ case view esMode es of
                  Command -> (blue, "C")
                  Editing -> (green, "E")
                  Selection -> case sMode $ fromJust $ esSelection es of
                    CharWise -> (yellow, "S")
                    LineWise -> (yellow, "SL")
                  Interactive  -> (magenta, dTag $ fromJust $ esDialog es)
                  Error -> (red, "Error")
            left = string attr $ printf "%s%s (%s)"
                                   (B.name buffer)
                                   (if view B.isModified buffer then "+" else " ")
                                   (case B._filePath buffer of
                                      Just path -> path
                                      Nothing   -> "no path")
            right = string attr $ printf "%d:%d %s %s"
                                    (B.pointColumn buffer)
                                    (B.pointLine buffer)
                                    options
                                    (B.encoding buffer)
            options :: String
            options = printf "%s%s%s"
              (if _optKeyboardLayout opts == KM.Dvorak then "🅓 " else "🅠 ")
              (if _optEnableSmartNewline opts then "➥ " else "")
              (if _optEnableSmartStartOfLine opts then "⭠" else "⭰")
            spacer = charFill attr ' ' (winW - imageWidth mode - imageWidth left - imageWidth right) 1
            attr = defAttr `withBackColor` rgbColor 200 200 200

displayPoint :: Int -> B.Buffer -> Editor ()
displayPoint off buffer = do
  vty <- ask

  -- Cursor must be shown after calling update?
  showCursor (outputIface vty)
  setCursorPos (outputIface vty) (off+x) y

  where x = B.pointColumn buffer
        y = B.pointLine buffer - B.winY (B.window buffer)


processEvent :: Event -> Editor ()
processEvent ev = do
  es <- get
  case view esMode es of
    -- TODO: Factor out the modes cases somehow.
    Command -> case KM.lookup' (esKeyPrefix es) ev commandModeKMs of
                 KM.Match cmd -> put (es { esKeyPrefix = Nothing }) >> cmd
                 KM.Prefix    -> put $ es { esKeyPrefix = Just ev }
                 KM.NoMatch   -> put $ es { esKeyPrefix = Nothing }

    Selection -> case KM.lookup' (esKeyPrefix es) ev [selectionModeKM] of
                   KM.Match cmd -> put (es { esKeyPrefix = Nothing }) >> cmd >> updateSelection
                   KM.Prefix    -> put $ es { esKeyPrefix = Just ev }
                   KM.NoMatch   -> put $ es { esKeyPrefix = Nothing }

    Editing -> case KM.lookup' (esKeyPrefix es) ev [editingKM] of
                 KM.Match cmd -> put (es { esKeyPrefix = Nothing }) >> cmd
                 KM.Prefix    -> put $ es { esKeyPrefix = Just ev }
                 KM.NoMatch   -> put (es { esKeyPrefix = Nothing }) >> selfInsert ev

    Interactive -> do
      let d = fromJust $ esDialog es
      case ev of
        EvKey (KChar 'q') [MCtrl] -> quit
        EvKey KEnter [] -> fromMaybe (return ()) (dFinish d) >> dialogHide >> commandMode
        EvKey KEsc   [] -> dialogHide >> gotoPosition (dPoint $ fromJust $ esDialog es) >> commandMode
        EvKey (KChar c) []  -> dialogInsertText [c] >> fromMaybe (return ()) (dUpdate d)
        EvKey KBS [] -> dialogBackspace >> fromMaybe (return ()) (dUpdate d)

        EvKey (KChar 't') [MCtrl] -> searchNext Forward
        EvKey (KChar 'c') [MCtrl] -> searchNext Backward
        otherwise -> return ()
      -- This is a bit of a hack. Because of the async nature of the interactive dialog,
      -- make sure the mode is set to Error when an error ocurred inside the dialog.
      es <- get
      when (isJust $ view esError es) $ switchMode Error

    -- Any input will do.
    Error -> assign esError Nothing >> commandMode

    where
      selfInsert :: Event -> Editor ()
      selfInsert (EvKey (KChar c) []) = insertText [c]
      selfInsert (EvKey KEnter []) = insertSmartNewline
      selfInsert _ = return ()

runEditor :: Editor ()
runEditor = do
  vty <- ask
  es <- get

  display es

  ev <- liftIO $ nextEvent vty
  put $ es { esLastEvent = Just ev }
  processEvent ev

  es <- get
  unless (esQuit es) runEditor

main :: IO ()
main = do
  (buffer:rest) <- getArgs >>= mapM loadFile
  vty <- standardIOConfig >>= mkVty
  execRWST runEditor vty $ EditorState { _esBufferZipper = B.BufferZipper [] buffer rest
                                       , esSelection     = Nothing
                                       , _esMode         = Command
                                       , esClipboard     = Nothing
                                       , _esOptions      = defaultOptions
                                       , esKeyPrefix     = Nothing
                                       , esLastEvent     = Nothing
                                       , esDialog        = Nothing
                                       , _esSearchContext = SearchContext 0 Forward
                                       , _esError        = Nothing
                                       , esQuit          = False
                                       }
  shutdown vty

  where loadFile f = catch (B.loadBuffer f) $ \e -> if isDoesNotExistError e then return B.newBuffer { B.name = f } else throw e


-- Utils.

fst' (x, _, _) = x
snd' (_, x, _) = x
trd' (_, _, x) = x
