{-# LANGUAGE TemplateHaskell #-}

module Buffer where

import           Control.Lens (makeLenses)
import           Data.Char (isSpace)
import           Data.Maybe (fromJust, fromMaybe)
import           System.FilePath (splitFileName)
import           Types (Syntax, Selections)
import qualified VtyUI as UI
import           Yi.Rope (YiString)
import qualified Yi.Rope as R
import           Prelude hiding (repeat)

type Point = Int

-- | Selection anchor.
type Anchor = (Point, Point, Point) -- (<line start index>, <cursor position>, <line end index>)

-- | A view into a buffer.
data Window = Window
  { winY      :: Int
  , winWidth  :: Int
  , winHeight :: Int
  }

--data BufferOptions = BufferOptions

data EditType = Insert
              | Delete
              deriving (Eq)

data Edit =
  Edit { ePoint :: Point
       , eEdit  :: EditType
       , eText  :: YiString
       }

data Buffer = Buffer
  { text        :: YiString
  , encoding    :: String
  , point       :: Point
  , lastColumn  :: Point
  , name        :: String
  , _filePath   :: Maybe FilePath
  , _isModified :: Bool
  , window      :: Window
  , syntax      :: Syntax
  , selections  :: Selections
  , lastInsertedText :: YiString
  , lastDeletedText  :: YiString
  , lastEditPoint    :: Maybe (Point, Point)
  -- ^ Start and end of the last edit.
  , mark        :: Point
  -- ^ User's custom mark.
  , inEdit      :: Bool
  , edits       :: [Edit]
  -- ^ Edits in the current editing run.
  , undoStack   :: [[Edit]]
  , redoStack   :: [[Edit]]
  }

newBuffer :: Buffer
newBuffer = Buffer
  { text        = R.empty
  , encoding    = "UTF-8"
  , point       = 0
  , lastColumn  = 0
  , name        = "Untitled"
  , _filePath   = Nothing
  , _isModified = False
  , window      = Window 0 (UI.rawScreenWidth - 2) (UI.rawScreenHeight - 2)
  , syntax      = []
  , selections  = []
  , lastInsertedText = R.empty
  , lastDeletedText  = R.empty
  , lastEditPoint    = Nothing
  , mark        = 0
  , inEdit      = False
  , edits       = []
  , undoStack   = []
  , redoStack   = []
  }


-- Utils.

pointMax :: Buffer -> Point
pointMax b = R.length $ text b

endOfLineAt :: Point -> YiString -> Point
endOfLineAt p txt = R.length left - newline
  where (left, right) = R.splitAtLine (pointLineAt p txt + 1) txt
        newline       = if right == R.empty then 0 else 1

endOfLine :: Buffer -> Point
endOfLine b = endOfLineAt (point b) (text b)

startOfLineAt :: Point -> YiString -> Point
startOfLineAt p txt = R.length $ fst $ R.splitAtLine (pointLineAt p txt) txt

startOfLine :: Buffer -> Point
startOfLine b = startOfLineAt (point b) (text b)

-- Doesn't count the newline character.
lengthOfLineAt :: Point -> YiString -> Point
lengthOfLineAt p txt = endOfLineAt p txt - startOfLineAt p txt

pointLineAt :: Point -> YiString -> Point
pointLineAt p txt = R.countNewLines $ R.take p txt

pointLine :: Buffer -> Int
pointLine b = pointLineAt (point b) (text b)

pointColumnAt :: Point -> YiString -> Int
pointColumnAt p txt = p - startOfLineAt p txt

pointColumn :: Buffer -> Int
pointColumn b = pointColumnAt (point b) (text b)

currentLine :: Buffer -> YiString
currentLine b = line `R.append` (R.fromString "\n")
  where line = R.takeWhile (/= '\n') $ snd $ R.splitAtLine (pointLine b) (text b)

-- | Return the current line's indentation in number of spaces.
-- | Doesn't work for tabs, yet.
lineIndent :: Buffer -> Int
lineIndent b = R.length indent
  where (_, right) = R.splitAtLine (pointLine b) (text b)
        indent     = R.takeWhile (\c -> c == ' ') right

repeat :: Int -> (Buffer -> Buffer) -> Buffer -> Buffer
repeat n action b
  | n <= 0    = b
  | otherwise = repeat (n-1) action (action b)

charAt p txt = let (_, rest) = R.splitAt p txt
               in fromMaybe '\0' $ R.head rest

-- Doesn't work when the delimiters are escaped or are part of syntax constructs.
-- (for example parentheses inside strings)
findPairAt :: Char -> Char -> Point -> YiString -> Maybe (Point, Point)
findPairAt left right from txt =
  let start = start' 0 from txt
      end   = end' 0 (from+1) txt
  in if start == Nothing || end == Nothing
     then Nothing
     else Just (fromJust start, fromJust end)
  where
    start' rdelims p txt | p < 0 || p >= R.length txt           = Nothing
                         | rdelims == 0 && charAt p txt == left = Just p
                         | rdelims > 0  && charAt p txt == left = start' (rdelims-1) (p-1) txt
                         | charAt p txt == right                = start' (rdelims+1) (p-1) txt
                         | otherwise                            = start' rdelims (p-1) txt
    end' ldelims p txt | p < 0 || p >= R.length txt            = Nothing
                       | ldelims == 0 && charAt p txt == right = Just (p+1)
                       | ldelims > 0  && charAt p txt == right = end' (ldelims-1) (p+1) txt
                       | charAt p txt == left                  = end' (ldelims+1) (p+1) txt
                       | otherwise                             = end' ldelims (p+1) txt

findPair :: Char -> Char -> Buffer -> Maybe (Point, Point)
findPair left right b = findPairAt left right (point b) (text b)


-- Movement.

adjustLastColumn :: Buffer -> Buffer
adjustLastColumn b = b { lastColumn = pointColumnAt (point b) (text b) }

moveRight :: Buffer -> Buffer
moveRight b = adjustLastColumn $ b { point = min (pointMax b) (point b + 1) }

moveLeft :: Buffer -> Buffer
moveLeft b = adjustLastColumn $ b { point = max 0 (point b - 1) }

moveDown :: Buffer -> Buffer
moveDown b = b { point = min (pointMax b) (p + min (lengthOfLineAt p (text b)) (lastColumn b)) }
  where p = endOfLine b + 1

moveUp :: Buffer -> Buffer
moveUp b = b { point = max 0 $ startOfLineAt (start-1) (text b) + min (lengthOfLineAt (start-1) (text b)) (lastColumn b) }
  where start = startOfLineAt (point b) (text b)

moveWordRight :: Buffer -> Buffer
moveWordRight b = adjustLastColumn $ b { point = min (pointMax b) (p + R.length w + s) }
  where txt = R.drop p $ text b
        p   = point b
        (w, _)         = R.break isSpace txt
        (spaces, rest) = R.span isSpace txt
        (w', _)        = R.break isSpace rest
        s   = if spaces == R.empty
              then 0
              else R.length spaces + R.length w'

moveWordLeft :: Buffer -> Buffer
moveWordLeft b = adjustLastColumn $ b { point = min (pointMax b) (p - R.length w - s) }
  where txt = R.reverse $ R.take p $ text b
        p   = point b
        (w, _)         = R.break isSpace txt
        (spaces, rest) = R.span isSpace txt
        (w', _)        = R.break isSpace rest
        s   = if spaces == R.empty
              then 0
              else R.length spaces + R.length w'

moveEndOfLine :: Buffer -> Buffer
moveEndOfLine b = adjustLastColumn $ b { point = endOfLine b }

moveStartOfLine :: Bool -> Buffer -> Buffer
moveStartOfLine smart b = adjustLastColumn $ b { point = p }
  where i = lineIndent b
        s = startOfLine b
        p = if smart
            then if point b == i + s
                 then s
                 else i + s
            else s

movePageDown :: Buffer -> Buffer
movePageDown b = repeat (winHeight $ window b) moveDown b

movePageUp :: Buffer -> Buffer
movePageUp b = repeat (winHeight $ window b) moveUp b

moveEndOfText :: Buffer -> Buffer
moveEndOfText b = adjustLastColumn $ b { point = pointMax b }

moveStartOfText :: Buffer -> Buffer
moveStartOfText b = adjustLastColumn $ b { point = 0 }

gotoPosition :: Int -> Buffer -> Buffer
gotoPosition p b = adjustLastColumn $ b { point = p' }
  where p' = max 0 $ min (pointMax b) p

gotoLine :: Int -> Buffer -> Buffer
gotoLine l b = gotoPosition (R.length $ fst $ R.splitAtLine l $ text b) b

leaveMark :: Buffer -> Buffer
leaveMark b = b { mark = point b }

-- Editing.

insertText :: YiString -> Buffer -> Buffer
insertText txt b = adjustLastColumn $
  b { text  = left `R.append` txt `R.append` right
    , point = p
    , lastInsertedText = txt
    , lastEditPoint    = Just (point b, p)
    , mark  = m
    }
  where (left, right) = R.splitAt (point b) (text b)
        p             = point b + R.length txt
        m             = if mark b >= point b then mark b + R.length txt else mark b

deleteText :: Int -> Buffer -> Buffer
deleteText n b = adjustLastColumn $
  b { text = left `R.append` right
    , lastDeletedText = R.take n rest
    , lastEditPoint   = Just (point b, point b)
    , mark = m
    }
  where (left, rest)     = R.splitAt (point b) (text b)
        (deleted, right) = R.splitAt n rest
        m                = if mark b >= point b then max (point b) (mark b - n) else mark b

maybeCreateUndo :: Buffer -> Buffer
maybeCreateUndo b = if inEdit b then id b else endEdit b

insertTextWithUndo :: YiString -> Buffer -> Buffer
insertTextWithUndo txt = maybeCreateUndo . createEdit Insert . insertText txt

deleteTextWithUndo :: Int -> Buffer -> Buffer
deleteTextWithUndo n = maybeCreateUndo . createEdit Delete . deleteText n

-- | Delete all whitespace around the point and leave only one space.
cleanupWhitespace :: Buffer -> Buffer
cleanupWhitespace b = deleteTextWithUndo spaces $ b { point = p }
  where (l, r) = R.splitAt (point b) (text b)
        left   = R.takeWhileEnd isSpace' l
        right  = R.takeWhile    isSpace' r
        spaces = R.length left + R.length right - 1
        p      = point b - R.length left
        isSpace' c = c `elem` " \t\r\v\f"


-- Redisplay.

adjustWindowAt :: Int -> Window -> Window
adjustWindowAt line win@(Window y w h)
  | line <  y   = Window line w h
  | line >= y+h = Window (line-h+1) w h
  | otherwise   = win

adjustWindow :: Buffer -> Buffer
adjustWindow b = b { window = adjustWindowAt (pointLine b) (window b) }

redisplay :: Buffer -> Buffer
redisplay b = b { window = win { winY = max 0 $ pointLine b - winHeight win `div` 2 } }
  where win = window b


-- IO.

loadBuffer :: FilePath -> IO Buffer
loadBuffer path = do
  res <- R.readFile path
  case res of
    Right (txt, converter) ->
      return Buffer { text = txt
                    , encoding = R.unCn converter
                    , point = 0
                    , lastColumn = 0
                    , name = snd $ splitFileName path
                    , _filePath = Just path
                    , _isModified = False
                    , window = Window 0 (UI.rawScreenWidth - 2) (UI.rawScreenHeight - 2)
                    , syntax = []
                    , selections = []
                    , lastInsertedText = R.empty
                    , lastDeletedText  = R.empty
                    , lastEditPoint    = Nothing
                    , mark = 0
                    , inEdit = False
                    , edits = []
                    , undoStack  = []
                    , redoStack  = []
                    }
    Left err -> return newBuffer -- TODO: Handle errors!

loadBuffers :: [FilePath] -> IO [Buffer]
loadBuffers [] = return [newBuffer]
loadBuffers paths = sequence $ fmap loadBuffer paths


-- Undo and redo.
-- | Undo currently works like this: The buffer keeps single edits in its 'edits' field.
-- | When an undo record is created, the 'edits' are simply put on an undo stack as one unit,
-- | which allows for grouping operations together. By default, every editing operation is created
-- | as a single item in 'edits' resulting in a single edit undo records. When one wants to group
-- | the operations together, startEdit and endEdit must be used to delimit the group of operations.
-- | Every startEdit has to have a matching endEdit, otherwise undo will break.
-- |
-- | Note: The implementation is probably quite slow and should have and upper limit on 'edits' and
-- | undo records, otherwise a lot of memory might be used. Also, because of the nature
-- | of the implementation, if a group is being recorded, the point of the undo record is set
-- | to where it was *after the last edit in that group*, which might result in a weirdly positioned
-- | point after undoing the group.

startEdit :: Buffer -> Buffer
startEdit b = b { inEdit = True }

endEdit :: Buffer -> Buffer
endEdit b = new { inEdit = False, edits = [] }
  where new = if null (edits b) then b else createUndo b

createEdit :: EditType -> Buffer -> Buffer
createEdit et b = b { edits = e:es }
  where e = Edit { ePoint = point b
                 , eEdit  = et
                 , eText  = txt
                 }
        es = edits b
        txt = if et == Insert then lastInsertedText b else lastDeletedText b

createUndo :: Buffer -> Buffer
createUndo b = b { undoStack = e:es, redoStack = [] }
  where e  = edits b
        es = undoStack b

-- FIXME: undo and redo should correctly set lastEditPoint in grouped undo records.
undo :: Buffer -> Buffer
undo b = case undoStack b of
  u:us -> new { undoStack = us, redoStack = u:redoStack b }
    where new = foldl apply b u
          apply b' (Edit p et txt) =
            case et of
              Insert -> deleteText (R.length txt) (gotoPosition (p - R.length txt) b')
              Delete -> gotoPosition p $ insertText txt (gotoPosition p b')
  [] -> b

redo :: Buffer -> Buffer
redo b = case redoStack b of
  u:us -> new { undoStack = u:undoStack b, redoStack = us }
    where new = foldr apply b u
          apply (Edit p et txt) b' =
            case et of
              Insert -> insertText txt (gotoPosition (p - R.length txt) b')
              Delete -> deleteText (R.length txt) (gotoPosition p b')
  [] -> b


-- Other. TODO sort

clearSelections :: Buffer -> Buffer
clearSelections b = b { selections = [] }



data BufferZipper = BufferZipper {
  bufBefore   :: [Buffer],
  _bufCurrent :: Buffer,
  bufAfter    :: [Buffer]
}

makeLenses ''Buffer
makeLenses ''BufferZipper

emptyBufferZipper :: BufferZipper
emptyBufferZipper = BufferZipper [] newBuffer []

openBuffer :: Buffer -> BufferZipper -> BufferZipper
openBuffer b (BufferZipper before cur after) = BufferZipper (cur:before) b after

nextBuffer :: BufferZipper -> BufferZipper
nextBuffer bz@(BufferZipper _ _ []) = bz
nextBuffer bz@(BufferZipper before cur (b:bs)) = BufferZipper (cur:before) b bs

previousBuffer :: BufferZipper -> BufferZipper
previousBuffer bz@(BufferZipper [] _ _) = bz
previousBuffer bz@(BufferZipper (b:bs) cur after) = BufferZipper bs b (cur:after)

closeBuffer :: BufferZipper -> BufferZipper
closeBuffer bz@(BufferZipper [] _ []) = bz
closeBuffer bz@(BufferZipper (b:bs) cur []) = BufferZipper bs b []
closeBuffer bz@(BufferZipper before _ (a:as)) = BufferZipper before a as
