module Cursor where

-- import Data.Char (isSpace)
-- import Data.List (elemIndex)
import qualified Data.Text as T

type Line = T.Text
type Anchor = (Int, Int, Int) -- (<line start index>, <cursor position>, <line end index>)

data LineZipper = LineZipper
  { lzBefore :: [Line]
  , lzAt     :: Line
  , lzAfter  :: [Line]
  }

data Cursor = Cursor
  { cLine       :: Int
  , cColumn     :: Int
  , cPosition   :: Int
  , cLastInsertedText :: T.Text
  , cLastDeletedText  :: T.Text
  -- ^ Text deleted during the last delete operation. Needed when performing undo.
  , cLineZipper :: LineZipper
  }

newCursor :: Cursor
newCursor = Cursor 0 0 0 T.empty T.empty $ LineZipper [] T.empty []

textToCursor :: T.Text -> Cursor
textToCursor text
  | text == T.empty = newCursor
  | otherwise = Cursor 0 0 0 T.empty T.empty $ LineZipper [] (head lines) (tail lines)
  where lines = T.lines text

cursorToLines :: Cursor -> [Line]
cursorToLines (Cursor _ _ _ _ _ (LineZipper before at after))
  = reverse before ++ [at] ++ after

cursorToText :: Cursor -> T.Text
cursorToText c
  = T.unlines $ cursorToLines c

-- Anchor utils.

cursorToAnchor :: Cursor -> Anchor
cursorToAnchor (Cursor _ col pos _ _ (LineZipper _ at _)) = (start, pos, end)
  where start = pos - col
        end   = pos + T.length at - col + 1 -- Newline.

-- Moving.

moveRight :: Cursor -> Cursor
moveRight c@(Cursor line col pos itxt dtxt lz)
  | col >= T.length (lzAt lz) = moveDown c
  | otherwise                 = Cursor line (col+1) (pos+1) itxt dtxt lz

moveLeft :: Cursor -> Cursor
moveLeft c@(Cursor line col pos itxt dtxt lz@(LineZipper before _ _))
  | col <= 0  = upCursor { cColumn   = if null before then 0 else T.length $ head before
                         , cPosition = if null before then 0 else pos-1
                         }
  | otherwise = Cursor line (col-1) (pos-1) itxt dtxt lz
  where upCursor = moveUp c

moveDown :: Cursor -> Cursor
moveDown c@(Cursor _ _ _ _ _ (LineZipper _ _ [])) = c
moveDown c@(Cursor line col pos _ _ (LineZipper before at (l:ls))) =
  c { cLine       = line+1
    , cColumn     = 0
    , cPosition   = pos + T.length at - col + 1 -- Newline.
    , cLineZipper = LineZipper (at:before) l ls
    }

moveUp c@(Cursor _ _ _ _ _ (LineZipper [] _ _)) = c
moveUp c@(Cursor line col pos _ _ (LineZipper (l:ls) at after)) =
  c { cLine = line-1
    , cColumn = 0
    , cPosition = pos - T.length l - col - 1 -- Newline.
    , cLineZipper = LineZipper ls l (at:after)
    }

onEndOfLine :: Cursor -> Bool
onEndOfLine c = cColumn c == T.length (lzAt $ cLineZipper c)

onStartOfLine :: Cursor -> Bool
onStartOfLine c = cColumn c == 0

onPosition :: Int -> Cursor -> Bool
onPosition p c = cPosition c == p

moveRightWhile :: (Cursor -> Bool) -> Cursor -> Cursor
moveRightWhile pred c
  | pred c    = moveRightWhile pred $ moveRight c
  | otherwise = c

moveLeftWhile :: (Cursor -> Bool) -> Cursor -> Cursor
moveLeftWhile pred c
  | pred c    = moveLeftWhile pred $ moveLeft c
  | otherwise = c

moveEndOfLine :: Cursor -> Cursor
moveEndOfLine = moveRightWhile (not . onEndOfLine)

moveStartOfLine :: Cursor -> Cursor
moveStartOfLine = moveLeftWhile (not . onStartOfLine)

-- TODO: Super Slow (TM).
moveEndOfText :: Cursor -> Cursor
moveEndOfText c = gotoPosition (T.length (cursorToText c) - 1) c

moveStartOfText :: Cursor -> Cursor
moveStartOfText c@(Cursor _ _ _ _ _ bc@(LineZipper [] at after)) =
  c { cLine = 0
    , cColumn = 0
    , cPosition = 0
    }
moveStartOfText c@(Cursor _ _ _ _ _ bc@(LineZipper before at after)) =
  c { cLine = 0
    , cColumn = 0
    , cPosition = 0
    , cLineZipper = LineZipper [] (last before) $ (reverse $ init before) ++ at:after
    }

-- TODO: Should check maximum buffer position!
gotoPosition :: Int -> Cursor -> Cursor
gotoPosition p c = if p > cPosition c
                   then moveRightWhile (not . onPosition p) c
                   else moveLeftWhile  (not . onPosition p) c

insertText :: T.Text -> Cursor -> Cursor
insertText str c@(Cursor line col pos itxt dtxt bc@(LineZipper before at after))
  | nlines == 0 = c
  | nlines == 1 && not endsWithNewline =
      Cursor line (col + strLen) (pos + strLen) str dtxt $ bc { lzAt = T.append (T.take col at) (T.append str $ T.drop col at) }
  | nlines == 1 && endsWithNewline =
      Cursor (line + 1) 0 (pos + strLen) str dtxt $ bc { lzBefore = T.append (T.take col at) (T.init str) : before
                                                       , lzAt = T.drop col at
                                                       }

  | nlines > 1 && not endsWithNewline =
      Cursor { cLine = line + nlines - 1
             , cColumn = T.length $ last ls
             , cPosition = pos + strLen
             , cLastInsertedText = str
             , cLastDeletedText = dtxt
             , cLineZipper =  bc { lzBefore = (drop 1 $ reverse $ tail ls) ++ [T.append (T.take col at) (head ls)] ++ before
                                 , lzAt = T.append (last ls) (T.drop col at)
                                 }
             }
  | nlines > 1 && endsWithNewline =
      Cursor { cLine = line + nlines
             , cColumn = 0
             , cPosition = pos + strLen
             , cLastInsertedText = str
             , cLastDeletedText = dtxt
             , cLineZipper =  bc { lzBefore = (reverse $ tail ls) ++ [T.append (T.take col at) (head ls)] ++ before
                                 , lzAt = T.drop col at
                                 }
             }
  where
    ls     = T.lines str
    nlines = length ls
    strLen = T.length str
    endsWithNewline = T.last str == '\n'


delete :: Cursor -> Cursor
delete c@(Cursor _ col pos _ _ bc@(LineZipper _ at [])) =
  c { cLineZipper = bc { lzAt = T.append (T.take col at) (T.drop (col+1) at) }
    , cLastDeletedText = T.take 1 $ T.drop col at
    }
delete c@(Cursor _ col pos _ _ bc@(LineZipper _ at (l:ls))) =
  c { cLineZipper = newContext
    , cLastDeletedText = newDeleted
    }
  where newContext = if col == T.length at
                     then bc { lzAt    = T.append at l,
                               lzAfter = ls
                             }
                     else bc { lzAt = T.append (T.take col at) (T.drop (col+1) at) }
        newDeleted = if col == T.length at
                     then T.pack "\n"
                     else T.take 1 $ T.drop col at

backspace :: Cursor -> Cursor
backspace c@(Cursor _ 0 _ _ _ (LineZipper [] _ _)) = c
backspace c@(Cursor line 0 pos itxt dtxt bc@(LineZipper (l:ls) at _)) =
  Cursor (line-1) (T.length l) (pos-1) itxt dtxt $ bc { lzBefore = ls,
                                                        lzAt = T.append l at
                                                      }
backspace c@(Cursor line col pos itxt dtxt bc@(LineZipper _ at _)) =
  Cursor line (col-1) (pos-1) itxt dtxt $ bc { lzAt = T.append (T.take (col-1) at) (T.drop col at) }


deleteEndOfLine :: Cursor -> Cursor
deleteEndOfLine c@(Cursor _ col _ _ _ bc@(LineZipper _ at _)) = c { cLineZipper = bc { lzAt = T.take col at }
                                                                  , cLastDeletedText = T.drop col at
                                                                  }

deleteStartOfLine :: Cursor -> Cursor
deleteStartOfLine c@(Cursor _ col pos _ _ bc@(LineZipper _ at _)) = c { cColumn = 0
                                                                      , cPosition = pos - col
                                                                      , cLineZipper = bc { lzAt = T.drop col at }
                                                                      , cLastDeletedText = T.take col at
                                                                      }
deleteLine :: Cursor -> Cursor
deleteLine c@(Cursor _ col pos _ _ bc@(LineZipper _ at [])) = c { cColumn = 0
                                                                , cPosition = pos - col
                                                                , cLineZipper = bc { lzAt = T.empty }
                                                                , cLastDeletedText = T.append at $ T.pack "\n"
                                                                }
deleteLine c@(Cursor _ col pos _ _ bc@(LineZipper _ at (l:ls))) =
  c { cColumn = 0
    , cPosition = pos - col
    , cLineZipper = bc { lzAt = l,
                         lzAfter = ls
                       }
    , cLastDeletedText = T.append at $ T.pack "\n"
    }
