-- UI backend for Vty
--

module VtyUI where

import           Data.Array
import qualified Data.Text    as T
import qualified Yi.Rope      as R
import qualified Graphics.Vty as Vty
import           System.IO.Unsafe (unsafePerformIO)
import           Text.Regex.PCRE
import           Types (Selections, Syntax)


-- | Terminal size.
-- A bit of a hack to get the terminal size.
region :: IO Vty.DisplayRegion
region = do
  vty <- Vty.standardIOConfig >>= Vty.mkVty
  r <- Vty.displayBounds $ Vty.outputIface vty
  Vty.shutdown vty
  return r

rawScreenWidth :: Int
rawScreenWidth = fst $ unsafePerformIO region

rawScreenHeight :: Int
rawScreenHeight = snd $ unsafePerformIO region


-- | Pre-drawing. Process the raw text to lines of (Char, Vty.Attr).
-- The syntax is rendered first, so the reverse video can be correctly applied
-- afterwards.
preDraw :: R.YiString              -- ^ The raw text.
        -> Int                     -- ^ Position of the first character in relation to the entire text.
        -> [(Int, Int)]            -- ^ Regions to highlight in reverse video (selection).
        -> [(Int, Int, Vty.Color)] -- ^ Regions to highlight in color (syntax).

        -> [[(Char, Vty.Attr)]]    -- ^ The pre-drawn lines of data to actually draw.
preDraw txt off regs syntax = lines' $ zip (R.toString txt) $ highlightRegions $ highlightSyntax $
                                       zip3 (R.toString txt) (repeat Vty.defAttr) [off..]
  where
    highlightRegions :: [(Char, Vty.Attr, Int)] -> [Vty.Attr]
    highlightRegions = map hr

    hr :: (Char, Vty.Attr, Int) -> Vty.Attr
    hr (_, attr, idx) = if (any (\(from, to) -> idx >= from && idx < to) regs)
                        then attr `Vty.withBackColor` Vty.rgbColor 88 151 184
                        else attr

    highlightSyntax :: [(Char, Vty.Attr, Int)] -> [(Char, Vty.Attr, Int)]
    highlightSyntax = map hs

    hs :: (Char, Vty.Attr, Int) -> (Char, Vty.Attr, Int)
    hs chr = foldl hs' chr syntax
    hs' id@(c, attr, idx) (from, to, color) = if idx >= from && idx < to
                                              then (c, attr `Vty.withForeColor` color, idx)
                                              else id

    lines' :: [(Char, Vty.Attr)] -> [[(Char, Vty.Attr)]]
    lines' [] = []
    lines' s  = case s' of
                  []          -> [l]
                  ((_,_):s'') -> l : lines' s''
                where
                  (l, s') = break ((== '\n') . fst) s


-- | Draw one pre-drawn line into an image.
drawLine :: Int -> Int -> [(Char, Vty.Attr)] -> Vty.Image
drawLine winW winH []   = Vty.charFill Vty.defAttr ' ' winW 1
drawLine winW winH line = Vty.horizCat (map colorChar line)
                          Vty.<|>
                          Vty.charFill Vty.defAttr ' ' (winW - length line) 1
  where colorChar (c, a) = Vty.char a c

-- -- | The main module entry point.
-- -- Turn raw input into an image.
-- drawText :: T.Text                  -- ^ The raw text.
--          -> [(Int, Int)]            -- ^ Regions to highlight in reverse video (selection).
--          -> [(Int, Int, Vty.Color)] -- ^ Regions to highlight in color (syntax).
--          -> Window                  -- ^ The window into the buffer.
--          -> Vty.Image               -- ^ The final image.
-- drawText txt highlights syntax win = foldl (Vty.<->) Vty.emptyImage $
--                                      map (drawLine win) $ preDraw txt highlights syntax win

drawText :: R.YiString
         -> Int -- ^ Position of the first character of the string (because the string had been clamped, it can't be computed here).
         -> Selections
         -> Syntax
         -> Int -- ^ Window width.
         -> Int -- ^ Window height.
         -> Vty.Image
drawText txt off selections syntax winW winH = foldl (Vty.<->) Vty.emptyImage $
                                               map (drawLine winW winH) $ preDraw txt off selections syntax

-- -- | Syntax specific helpers.
-- syntaxFromText :: String -> Syntax
-- syntaxFromText text = keywords ++ strings ++ comments
--   where
--     keywords = map (tupleToTriple Vty.green) $ map (!0) $ matchAll reKeywords text
--     strings  = map (tupleToTriple Vty.yellow) $ map (!0) $ matchAll reStrings text
--     comments = map (tupleToTriple Vty.blue) $ map (!0) $ matchAll reComments text
--     reKeywords = makeRegex "\\b(import|where)\\b" :: Regex
--     reStrings  = makeRegex "\"[^\"\\\\]*(\\\\.[^\"\\\\]*)*\"" :: Regex
--     reComments = makeRegex "--.*$" :: Regex
--     tupleToTriple x (a, b) = (a, a+b, x)

-- syntaxFromText' :: T.Text -> Syntax
-- syntaxFromText' = syntaxFromText . T.unpack

